@@ -3,7 +3,7 @@
 #
 # library for communicating with an isc dhcp server over the omapi protocol
 #
-# Copyright (C) 2010-2011 Cygnus Networks GmbH
+# Copyright (C) 2010-2013 Cygnus Networks GmbH
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -40,23 +40,26 @@ For an example see http://code.google.com/p/pypureomapi/wiki/Example.
 __author__      = "Helmut Grohne, Torge Szczepanek"
 __copyright__   = "Cygnus Networks GmbH"
 __licence__     = "GPL-3"
-__version__     = "0.2"
+__version__     = "0.3"
 __maintainer__  = "Torge Szczepanek"
 __email__       = "info@cygnusnetworks.de"
 
 __all__ = []
 
+import binascii
 import struct
 import hmac
+import io
+import logging
 import socket
 import random
 import operator
-import sys
 try:
-	from cStringIO import StringIO
-except ImportError:
-	from StringIO import StringIO
+	basestring
+except NameError:
+	basestring = str
 
+logger = logging.getLogger("pypureomapi")
 sysrand = random.SystemRandom()
 
 __all__.extend("OMAPI_OP_OPEN OMAPI_OP_REFRESH OMAPI_OP_UPDATE".split())
@@ -93,21 +96,34 @@ class OmapiErrorNotFound(OmapiError):
 	def __init__(self):
 		OmapiError.__init__(self, "not found")
 
-class OutBuffer:
+class OutBuffer(object):
 	"""Helper class for constructing network packets."""
 	sizelimit = 65536
 	def __init__(self):
-		self.buff = StringIO()
+		self.buff = io.BytesIO()
+
+	def __len__(self):
+		"""Return the number of bytes in the buffer.
+		@rtype: int
+		"""
+		# On Py2.7 tell returns long, but __len__ is required to return int.
+		return int(self.buff.tell())
 
 	def add(self, data):
 		"""
-		@type data: str
+		>>> ob = OutBuffer().add(OutBuffer.sizelimit * b"x")
+		>>> ob.add(b"y") # doctest: +ELLIPSIS
+		Traceback (most recent call last):
+		...
+		OmapiSizeLimitError: ...
+
+		@type data: bytes
 		@returns: self
 		@raises OmapiSizeLimitError:
 		"""
-		self.buff.write(data)
-		if self.buff.tell() > self.sizelimit:
+		if len(self) + len(data) > self.sizelimit:
 			raise OmapiSizeLimitError()
+		self.buff.write(data)
 		return self
 
 	def add_net32int(self, integer):
@@ -134,10 +150,11 @@ class OutBuffer:
 
 	def add_net32string(self, string):
 		"""
-		>>> OutBuffer().add_net32string("x").getvalue()
-		'\\x00\\x00\\x00\\x01x'
+		>>> r = b'\\x00\\x00\\x00\\x01x'
+		>>> OutBuffer().add_net32string(b"x").getvalue() == r
+		True
 
-		@type string: str
+		@type string: bytes
 		@param string: maximum length must fit in a 32bit integer
 		@returns: self
 		@raises OmapiSizeLimitError:
@@ -148,10 +165,10 @@ class OutBuffer:
 
 	def add_net16string(self, string):
 		"""
-		>>> OutBuffer().add_net16string("x").getvalue()
-		'\\x00\\x01x'
+		>>> OutBuffer().add_net16string(b"x").getvalue() == b'\\x00\\x01x'
+		True
 
-		@type string: str
+		@type string: bytes
 		@param string: maximum length must fit in a 16bit integer
 		@returns: self
 		@raises OmapiSizeLimitError:
@@ -162,10 +179,11 @@ class OutBuffer:
 
 	def add_bindict(self, items):
 		"""
-		>>> OutBuffer().add_bindict(dict(foo="bar")).getvalue()
-		'\\x00\\x03foo\\x00\\x00\\x00\\x03bar\\x00\\x00'
+		>>> r = b'\\x00\\x03foo\\x00\\x00\\x00\\x03bar\\x00\\x00'
+		>>> OutBuffer().add_bindict({b"foo": b"bar"}).getvalue() == r
+		True
 
-		@type items: [(str, str)] or {str: str}
+		@type items: [(bytes, bytes)] or {bytes: bytes}
 		@returns: self
 		@raises OmapiSizeLimitError:
 		"""
@@ -173,29 +191,88 @@ class OutBuffer:
 			items = items.items()
 		for key, value in items:
 			self.add_net16string(key).add_net32string(value)
-		return self.add("\x00\x00") # end marker
+		return self.add(b"\x00\x00") # end marker
 
 	def getvalue(self):
 		"""
-		>>> OutBuffer().add("sp").add("am").getvalue()
-		'spam'
+		>>> OutBuffer().add(b"sp").add(b"am").getvalue() == b"spam"
+		True
 
-		@rtype: str
+		@rtype: bytes
 		"""
 		return self.buff.getvalue()
 
 	def consume(self, length):
 		"""
-		>>> OutBuffer().add("spam").consume(2).getvalue()
-		'am'
+		>>> OutBuffer().add(b"spam").consume(2).getvalue() == b"am"
+		True
 
 		@type length: int
 		@returns: self
 		"""
-		self.buff = StringIO(self.getvalue()[length:])
+		self.buff = io.BytesIO(self.getvalue()[length:])
 		return self
 
-class OmapiAuthenticatorBase:
+class OmapiStartupMessage(object):
+	"""Class describing the protocol negotiation messages.
+
+	>>> s = OmapiStartupMessage().as_string()
+	>>> s == b"\\0\\0\\0\\x64\\0\\0\\0\\x18"
+	True
+	>>> next(InBuffer(s).parse_startup_message()).validate()
+	>>> OmapiStartupMessage(42).validate()
+	Traceback (most recent call last):
+	...
+	OmapiError: protocol mismatch
+	"""
+	implemented_protocol_version = 100
+	implemented_header_size = 4 * 6
+
+	def __init__(self, protocol_version=None, header_size=None):
+		"""
+		@type protocol_version: int or None
+		@type header_size: int or None
+		"""
+		if protocol_version is None:
+			protocol_version = self.implemented_protocol_version
+		if header_size is None:
+			header_size = self.implemented_header_size
+		self.protocol_version = protocol_version
+		self.header_size = header_size
+
+	def validate(self):
+		"""Checks whether this OmapiStartupMessage matches the implementation.
+		@raises OmapiError:
+		"""
+		if self.implemented_protocol_version != self.protocol_version:
+			raise OmapiError("protocol mismatch")
+		if self.implemented_header_size != self.header_size:
+			raise OmapiError("header size mismatch")
+
+	def as_string(self):
+		"""
+		@rtype: bytes
+		"""
+		ret = OutBuffer()
+		self.serialize(ret)
+		return ret.getvalue()
+
+	def serialize(self, outbuffer):
+		"""Serialize this OmapiStartupMessage to the given outbuffer.
+		@type outbuffer: OutBuffer
+		"""
+		outbuffer.add_net32int(self.protocol_version)
+		outbuffer.add_net32int(self.header_size)
+
+	def dump_oneline(self):
+		"""
+		@rtype: str
+		@returns: a human readable representation in one line
+		"""
+		return "protocol_version=%d header_size=%d" % (self.protocol_version,
+				self.header_size)
+
+class OmapiAuthenticatorBase(object):
 	"""Base class for OMAPI authenticators.
 	@cvar authlen: is the length of a signature as returned by the sign method
 	@type authlen: int
@@ -211,14 +288,14 @@ class OmapiAuthenticatorBase:
 		pass
 	def auth_object(self):
 		"""
-		@rtype: {str: str}
+		@rtype: {bytes: bytes}
 		@returns: object part of an omapi authentication message
 		"""
 		raise NotImplementedError
 	def sign(self, message):
 		"""
-		@type message: str
-		@rtype: str
+		@type message: bytes
+		@rtype: bytes
 		@returns: a signature of length self.authlen
 		"""
 		raise NotImplementedError()
@@ -231,39 +308,39 @@ class OmapiNullAuthenticator(OmapiAuthenticatorBase):
 	def auth_object(self):
 		return {}
 	def sign(self, _):
-		return ""
+		return b""
 
 class OmapiHMACMD5Authenticator(OmapiAuthenticatorBase):
 	authlen = 16
-	algorithm = "hmac-md5.SIG-ALG.REG.INT."
+	algorithm = b"hmac-md5.SIG-ALG.REG.INT."
 	def __init__(self, user, key):
 		"""
-		@type user: str
-		@type key: str
+		@type user: bytes
+		@type key: bytes
 		@param key: base64 encoded key
 		@raises binascii.Error: for bad base64 encoding
 		"""
 		OmapiAuthenticatorBase.__init__(self)
 		self.user = user
-		self.key = key.decode("base64")
+		self.key = binascii.a2b_base64(key)
 
 	def auth_object(self):
-		return dict(name=self.user, algorithm=self.algorithm)
+		return {b"name": self.user, b"algorithm": self.algorithm}
 
 	def sign(self, message):
 		"""
 		>>> authlen = OmapiHMACMD5Authenticator.authlen
-		>>> len(OmapiHMACMD5Authenticator("foo", 16*"x").sign("baz")) == authlen
+		>>> len(OmapiHMACMD5Authenticator(b"foo", 16*b"x").sign(b"baz")) == authlen
 		True
 
-		@type message: str
-		@rtype: str
+		@type message: bytes
+		@rtype: bytes
 		@returns: a signature of length self.authlen
 		"""
 		return hmac.HMAC(self.key, message).digest()
 
 __all__.append("OmapiMessage")
-class OmapiMessage:
+class OmapiMessage(object):
 	"""
 	@type authid: int
 	@ivar authid: The id of the message authenticator.
@@ -276,16 +353,16 @@ class OmapiMessage:
 	@ivar tid: Transmission identifier.
 	@type rid: int
 	@ivar rid: Receive identifier (of a response is the tid of the request).
-	@type message: [(str, str)]
+	@type message: [(bytes, bytes)]
 	@ivar message: A list of (key, value) pairs.
-	@type obj: [(str, str)]
+	@type obj: [(bytes, bytes)]
 	@ivar obj: A list of (key, value) pairs.
-	@type signature: str
+	@type signature: bytes
 	@ivar signature: A signature on this message as generated by an
 			authenticator.
 	"""
 	def __init__(self, authid=0, opcode=0, handle=0, tid=0, rid=0,
-			message=None, obj=None, signature=""):
+			message=None, obj=None, signature=b""):
 		"""
 		Construct an OmapiMessage from the given fields. No error
 		checking is performed.
@@ -296,8 +373,8 @@ class OmapiMessage:
 		@type tid: int
 		@param tid: The special value -1 causes a tid to be generated randomly.
 		@type rid: int
-		@type message: [(str, str)]
-		@type obj: [(str, str)]
+		@type message: [(bytes, bytes)]
+		@type obj: [(bytes, bytes)]
 		@type signature: str
 		@rtype: OmapiMessage
 		"""
@@ -318,27 +395,35 @@ class OmapiMessage:
 		"""
 		self.tid = sysrand.randrange(0, 1<<32)
 
+	def serialize(self, outbuffer, forsigning=False):
+		"""
+		@type outbuffer: OutBuffer
+		@type forsigning: bool
+		@raises OmapiSizeLimitError:
+		"""
+		if not forsigning:
+			outbuffer.add_net32int(self.authid)
+		outbuffer.add_net32int(len(self.signature))
+		outbuffer.add_net32int(self.opcode)
+		outbuffer.add_net32int(self.handle)
+		outbuffer.add_net32int(self.tid)
+		outbuffer.add_net32int(self.rid)
+		outbuffer.add_bindict(self.message)
+		outbuffer.add_bindict(self.obj)
+		if not forsigning:
+			outbuffer.add(self.signature)
+
 	def as_string(self, forsigning=False):
 		"""
 		>>> len(OmapiMessage().as_string(True)) >= 24
 		True
 
 		@type forsigning: bool
-		@rtype: str
+		@rtype: bytes
 		@raises OmapiSizeLimitError:
 		"""
 		ret = OutBuffer()
-		if not forsigning:
-			ret.add_net32int(self.authid)
-		ret.add_net32int(len(self.signature))
-		ret.add_net32int(self.opcode)
-		ret.add_net32int(self.handle)
-		ret.add_net32int(self.tid)
-		ret.add_net32int(self.rid)
-		ret.add_bindict(self.message)
-		ret.add_bindict(self.obj)
-		if not forsigning:
-			ret.add(self.signature)
+		self.serialize(ret, forsigning)
 		return ret.getvalue()
 
 	def sign(self, authenticator):
@@ -346,12 +431,26 @@ class OmapiMessage:
 		@type authenticator: OmapiAuthenticatorBase
 		"""
 		self.authid = authenticator.authid
-		self.signature = "\0" * authenticator.authlen # provide authlen
+		self.signature = b"\0" * authenticator.authlen # provide authlen
 		self.signature = authenticator.sign(self.as_string(forsigning=True))
 		assert len(self.signature) == authenticator.authlen
 
 	def verify(self, authenticators):
 		"""Verify this OMAPI message.
+
+		>>> a1 = OmapiHMACMD5Authenticator(b"egg", b"spam")
+		>>> a2 = OmapiHMACMD5Authenticator(b"egg", b"tomatoes")
+		>>> a1.authid = a2.authid = 5
+		>>> m = OmapiMessage.open(b"host")
+		>>> m.verify({a1.authid: a1})
+		False
+		>>> m.sign(a1)
+		>>> m.verify({a1.authid: a1})
+		True
+		>>> m.sign(a2)
+		>>> m.verify({a1.authid: a1})
+		False
+
 		@type authenticators: {int: OmapiAuthenticatorBase}
 		@rtype: bool
 		"""
@@ -365,10 +464,18 @@ class OmapiMessage:
 	@classmethod
 	def open(cls, typename):
 		"""Create an OMAPI open message with given typename.
-		@type typename: str
+		@type typename: bytes
 		@rtype: OmapiMessage
 		"""
-		return cls(opcode=OMAPI_OP_OPEN, message=[("type", typename)], tid=-1)
+		return cls(opcode=OMAPI_OP_OPEN, message=[(b"type", typename)], tid=-1)
+
+	@classmethod
+	def update(cls, handle):
+		"""Create an OMAPI update message for the given handle.
+		@type handle: int
+		@rtype: OmapiMessage
+		"""
+		return cls(opcode=OMAPI_OP_UPDATE, handle=handle, tid=-1)
 
 	@classmethod
 	def delete(cls, handle):
@@ -387,7 +494,7 @@ class OmapiMessage:
 
 	def update_object(self, update):
 		"""
-		@type update: {str: str}
+		@type update: {bytes: bytes}
 		"""
 		self.obj = [(key, value) for key, value in self.obj
 					if key not in update]
@@ -409,6 +516,17 @@ class OmapiMessage:
 				"obj:\t\t%r\n" % self.obj,
 				"signature:\t%r\n" % self.signature))
 
+	def dump_oneline(self):
+		"""
+		@rtype: str
+		@returns: a barely human readable representation in one line
+		"""
+		return ("authid=%d authlen=%d opcode=%s handle=%d tid=%d rid=%d " +
+				"message=%r obj=%r signature=%r") % (self.authid,
+						len(self.signature), repr_opcode(self.opcode),
+						self.handle, self.tid, self.rid, self.message,
+						self.obj, self.signature)
+
 def parse_map(filterfun, parser):
 	"""Creates a new parser that passes the result of the given parser through
 	the given filterfun.
@@ -447,22 +565,22 @@ def parse_chain(*args):
 				break
 	yield tuple(items)
 
-class InBuffer:
+class InBuffer(object):
 	sizelimit = 65536
-	def __init__(self, initial=""):
+	def __init__(self, initial=b""):
 		"""
-		@type initial: str
+		@type initial: bytes
 		@param initial: initial value of the buffer
 		@raises OmapiSizeLimitError:
 		"""
-		self.buff = ""
+		self.buff = b""
 		self.totalsize = 0
 		if initial:
 			self.feed(initial)
 
 	def feed(self, data):
 		"""
-		@type data: str
+		@type data: bytes
 		@returns: self
 		@raises OmapiSizeLimitError:
 		"""
@@ -491,7 +609,7 @@ class InBuffer:
 
 	def parse_net16int(self):
 		"""
-		>>> hex(next(InBuffer("\\x01\\x02").parse_net16int()))
+		>>> hex(next(InBuffer(b"\\x01\\x02").parse_net16int()))
 		'0x102'
 		"""
 		return parse_map(lambda data: struct.unpack("!H", data)[0],
@@ -499,7 +617,7 @@ class InBuffer:
 
 	def parse_net32int(self):
 		"""
-		>>> hex(next(InBuffer("\\x01\\0\\0\\x02").parse_net32int()))
+		>>> hex(int(next(InBuffer(b"\\x01\\0\\0\\x02").parse_net32int())))
 		'0x1000002'
 		"""
 		return parse_map(lambda data: struct.unpack("!L", data)[0],
@@ -507,25 +625,25 @@ class InBuffer:
 
 	def parse_net16string(self):
 		"""
-		>>> next(InBuffer("\\0\\x03eggs").parse_net16string())
-		'egg'
+		>>> next(InBuffer(b"\\0\\x03eggs").parse_net16string()) == b'egg'
+		True
 		"""
 		return parse_map(operator.itemgetter(1),
 				parse_chain(self.parse_net16int, self.parse_fixedbuffer))
 
 	def parse_net32string(self):
 		"""
-		>>> next(InBuffer("\\0\\0\\0\\x03eggs").parse_net32string())
-		'egg'
+		>>> next(InBuffer(b"\\0\\0\\0\\x03eggs").parse_net32string()) == b'egg'
+		True
 		"""
 		return parse_map(operator.itemgetter(1),
 				parse_chain(self.parse_net32int, self.parse_fixedbuffer))
 
 	def parse_bindict(self):
 		"""
-		>>> d = "\\0\\x01a\\0\\0\\0\\x01b\\0\\0spam"
-		>>> next(InBuffer(d).parse_bindict())
-		[('a', 'b')]
+		>>> d = b"\\0\\x01a\\0\\0\\0\\x01b\\0\\0spam"
+		>>> next(InBuffer(d).parse_bindict()) == [(b'a', b'b')]
+		True
 		"""
 		entries = []
 		try:
@@ -549,10 +667,17 @@ class InBuffer:
 			yield entries
 
 	def parse_startup_message(self):
-		# results in (version, headersize)
-		return parse_chain(self.parse_net32int, lambda _: self.parse_net32int())
+		"""results in an OmapiStartupMessage
+
+		>>> d = b"\\0\\0\\0\\x64\\0\\0\\0\\x18"
+		>>> next(InBuffer(d).parse_startup_message()).validate()
+		"""
+		return parse_map(lambda args: OmapiStartupMessage(*args),
+				parse_chain(self.parse_net32int,
+					lambda _: self.parse_net32int()))
 
 	def parse_message(self):
+		"""results in an OmapiMessage"""
 		parser = parse_chain(self.parse_net32int, # authid
 				lambda *_: self.parse_net32int(), # authlen
 				lambda *_: self.parse_net32int(), # opcode
@@ -565,6 +690,17 @@ class InBuffer:
 		return parse_map(lambda args: # skip authlen in args:
 				OmapiMessage(*(args[0:1] + args[2:])), parser)
 
+if isinstance(bytes(b"x")[0], int):
+	def bytes_to_int_seq(b):
+		return b
+	int_seq_to_bytes = bytes # raises ValueError
+else:
+	def bytes_to_int_seq(b):
+		return map(ord, b)
+	def int_seq_to_bytes(s):
+		return "".join(map(chr, s)) # raises ValueError
+
+__all__.append("pack_ip")
 def pack_ip(ipstr):
 	"""Converts an ip address given in dotted notation to a four byte
 	string in network byte order.
@@ -577,7 +713,7 @@ def pack_ip(ipstr):
 	ValueError: given ip address has an invalid number of dots
 
 	@type ipstr: str
-	@rtype: str
+	@rtype: bytes
 	@raises ValueError: for badly formatted ip addresses
 	"""
 	if not isinstance(ipstr, basestring):
@@ -586,36 +722,37 @@ def pack_ip(ipstr):
 	if len(parts) != 4:
 		raise ValueError("given ip address has an invalid number of dots")
 	parts = map(int, parts) # raises ValueError
-	parts = map(chr, parts) # raises ValueError
-	return "".join(parts) # network byte order
+	return int_seq_to_bytes(parts) # raises ValueError
 
+__all__.append("unpack_ip")
 def unpack_ip(fourbytes):
 	"""Converts an ip address given in a four byte string in network
 	byte order to a string in dotted notation.
 
-	>>> unpack_ip("dead")
+	>>> unpack_ip(b"dead")
 	'100.101.97.100'
-	>>> unpack_ip("alive")
+	>>> unpack_ip(b"alive")
 	Traceback (most recent call last):
 	...
 	ValueError: given buffer is not exactly four bytes long
 
-	@type fourbytes: str
+	@type fourbytes: bytes
 	@rtype: str
 	@raises ValueError: for bad input
 	"""
-	if not isinstance(fourbytes, str):
+	if not isinstance(fourbytes, bytes):
 		raise ValueError("given buffer is not a string")
 	if len(fourbytes) != 4:
 		raise ValueError("given buffer is not exactly four bytes long")
-	return ".".join(map(str, map(ord, fourbytes)))
+	return ".".join(map(str, bytes_to_int_seq(fourbytes)))
 
+__all__.append("pack_mac")
 def pack_mac(macstr):
 	"""Converts a mac address given in colon delimited notation to a
 	six byte string in network byte order.
 
-	>>> pack_mac("30:31:32:33:34:35")
-	'012345'
+	>>> pack_mac("30:31:32:33:34:35") == b'012345'
+	True
 	>>> pack_mac("bad")
 	Traceback (most recent call last):
 	...
@@ -623,7 +760,7 @@ def pack_mac(macstr):
 
 
 	@type macstr: str
-	@rtype: str
+	@rtype: bytes
 	@raises ValueError: for badly formatted mac addresses
 	"""
 	if not isinstance(macstr, basestring):
@@ -632,41 +769,44 @@ def pack_mac(macstr):
 	if len(parts) != 6:
 		raise ValueError("given mac addresses has an invalid number of colons")
 	parts = [int(part, 16) for part in parts] # raises ValueError
-	parts = map(chr, parts) # raises ValueError
-	return "".join(parts) # network byte order
+	return int_seq_to_bytes(parts) # raises ValueError
 
+__all__.append("unpack_mac")
 def unpack_mac(sixbytes):
 	"""Converts a mac address given in a six byte string in network
 	byte order to a string in colon delimited notation.
 
-	>>> unpack_mac("012345")
+	>>> unpack_mac(b"012345")
 	'30:31:32:33:34:35'
-	>>> unpack_mac("bad")
+	>>> unpack_mac(b"bad")
 	Traceback (most recent call last):
 	...
 	ValueError: given buffer is not exactly six bytes long
 
-	@type sixbytes: str
+	@type sixbytes: bytes
 	@rtype: str
 	@raises ValueError: for bad input
 	"""
-	if not isinstance(sixbytes, str):
+	if not isinstance(sixbytes, bytes):
 		raise ValueError("given buffer is not a string")
 	if len(sixbytes) != 6:
 		raise ValueError("given buffer is not exactly six bytes long")
-	return ":".join(map("%2.2x".__mod__, map(ord, sixbytes)))
+	return ":".join(map("%2.2x".__mod__, bytes_to_int_seq(sixbytes)))
+
+class lazy_str(object):
+	def __init__(self, function):
+		self.function = function
+	def __str__(self):
+		return self.function()
 
 __all__.append("Omapi")
-class Omapi:
-	protocol_version = 100
-
-	def __init__(self, hostname, port, username=None, key=None, debug=False):
+class Omapi(object):
+	def __init__(self, hostname, port, username=None, key=None):
 		"""
 		@type hostname: str
 		@type port: int
-		@type username: str or None
-		@type key: str or None
-		@type debug: bool
+		@type username: bytes or None
+		@type key: bytes or None
 		@param key: if given, it must be base64 encoded
 		@raises binascii.Error: for bad base64 encoding
 		@raises socket.error:
@@ -676,7 +816,6 @@ class Omapi:
 		self.port = port
 		self.authenticators = {0: OmapiNullAuthenticator()}
 		self.defauth = 0
-		self.debug = debug
 
 		newauth = None
 		if username is not None and key is not None:
@@ -709,6 +848,7 @@ class Omapi:
 		"""Receive up to length bytes of data from the connection.
 
 		@type length: int
+		@rtype: bytes
 		@raises OmapiError: if not connected
 		@raises socket.error:
 		"""
@@ -722,7 +862,7 @@ class Omapi:
 	def send_conn(self, data):
 		"""Send all of data to the connection.
 
-		@type data: str
+		@type data: bytes
 		@raises OmapiError: if not connected
 		@raises socket.error:
 		"""
@@ -754,10 +894,10 @@ class Omapi:
 		@raises socket.error:
 		"""
 		self.check_connected()
-		buff = OutBuffer()
-		buff.add_net32int(self.protocol_version)
-		buff.add_net32int(4*6) # header size
-		self.send_conn(buff.getvalue())
+		message = OmapiStartupMessage()
+		logger.debug("sending omapi startup message %s",
+				lazy_str(message.dump_oneline))
+		self.send_conn(message.as_string())
 
 	def recv_protocol_initialization(self):
 		"""
@@ -769,13 +909,13 @@ class Omapi:
 				self.fill_inbuffer()
 			else:
 				self.inbuffer.resetsize()
-				protocol_version, header_size = result
-				if protocol_version != self.protocol_version:
+				logger.debug("received omapi startup message %s",
+						lazy_str(result.dump_oneline))
+				try:
+					result.validate()
+				except OmapiError:
 					self.close()
-					raise OmapiError("protocol mismatch")
-				if header_size != 4*6:
-					self.close()
-					raise OmapiError("header size mismatch")
+					raise
 
 	def receive_message(self):
 		"""Read the next message from the connection.
@@ -788,12 +928,10 @@ class Omapi:
 				self.fill_inbuffer()
 			else:
 				self.inbuffer.resetsize()
+				logger.debug("received %s", lazy_str(message.dump_oneline))
 				if not message.verify(self.authenticators):
 					self.close()
 					raise OmapiError("bad omapi message signature")
-				if self.debug:
-					print "debug recv"
-					sys.stdout.write(message.dump())
 				return message
 
 	def receive_response(self, message, insecure=False):
@@ -825,9 +963,7 @@ class Omapi:
 		self.check_connected()
 		if sign:
 			message.sign(self.authenticators[self.defauth])
-		if self.debug:
-			print "debug send"
-			sys.stdout.write(message.dump())
+		logger.debug("sending %s", lazy_str(message.dump_oneline))
 		self.send_conn(message.as_string())
 
 	def query_server(self, message):
@@ -847,7 +983,7 @@ class Omapi:
 		@raises OmapiError:
 		@raises socket.error:
 		"""
-		msg = OmapiMessage.open("authenticator")
+		msg = OmapiMessage.open(b"authenticator")
 		msg.update_object(authenticator.auth_object())
 		response = self.query_server(msg)
 		if response.opcode != OMAPI_OP_UPDATE:
@@ -858,35 +994,38 @@ class Omapi:
 		self.authenticators[authid] = authenticator
 		authenticator.authid = authid
 		self.defauth = authid
+		logger.debug("successfully initialized default authid %d", authid)
 
 	def add_host(self, ip, mac):
-		"""
+		"""Create a host object with given ip address and and mac address.
+
 		@type ip: str
 		@type mac: str
 		@raises ValueError:
 		@raises OmapiError:
 		@raises socket.error:
 		"""
-		msg = OmapiMessage.open("host")
-		msg.message.append(("create", struct.pack("!I", 1)))
-		msg.message.append(("exclusive", struct.pack("!I", 1)))
-		msg.obj.append(("hardware-address", pack_mac(mac)))
-		msg.obj.append(("hardware-type", struct.pack("!I", 1)))
-		msg.obj.append(("ip-address", pack_ip(ip)))
+		msg = OmapiMessage.open(b"host")
+		msg.message.append((b"create", struct.pack("!I", 1)))
+		msg.message.append((b"exclusive", struct.pack("!I", 1)))
+		msg.obj.append((b"hardware-address", pack_mac(mac)))
+		msg.obj.append((b"hardware-type", struct.pack("!I", 1)))
+		msg.obj.append((b"ip-address", pack_ip(ip)))
 		response = self.query_server(msg)
 		if response.opcode != OMAPI_OP_UPDATE:
 			raise OmapiError("add failed")
 
 	def del_host(self, mac):
-		"""
+		"""Delete a host object with with given mac address.
+
 		@type mac: str
 		@raises ValueError:
 		@raises OmapiError:
 		@raises socket.error:
 		"""
-		msg = OmapiMessage.open("host")
-		msg.obj.append(("hardware-address", pack_mac(mac)))
-		msg.obj.append(("hardware-type", struct.pack("!I", 1)))
+		msg = OmapiMessage.open(b"host")
+		msg.obj.append((b"hardware-address", pack_mac(mac)))
+		msg.obj.append((b"hardware-type", struct.pack("!I", 1)))
 		response = self.query_server(msg)
 		if response.opcode != OMAPI_OP_UPDATE:
 			raise OmapiErrorNotFound()
@@ -897,38 +1036,46 @@ class Omapi:
 			raise OmapiError("delete failed")
 
 	def lookup_ip(self, mac):
-		"""
+		"""Look for a lease object with given mac address and return the
+		assigned ip address.
+
 		@type mac: str
 		@rtype: str or None
 		@raises ValueError:
 		@raises OmapiError:
+		@raises OmapiErrorNotFound: if no lease object with the given mac
+				address could be found or the object lacks an ip address
 		@raises socket.error:
 		"""
-		msg = OmapiMessage.open("lease")
-		msg.obj.append(("hardware-address", pack_mac(mac)))
+		msg = OmapiMessage.open(b"lease")
+		msg.obj.append((b"hardware-address", pack_mac(mac)))
 		response = self.query_server(msg)
 		if response.opcode != OMAPI_OP_UPDATE:
 			raise OmapiErrorNotFound()
 		try:
-			return unpack_ip(dict(response.obj)["ip-address"])
+			return unpack_ip(dict(response.obj)[b"ip-address"])
 		except KeyError: # ip-address
 			raise OmapiErrorNotFound()
 
 	def lookup_mac(self, ip):
-		"""
+		"""Look up a lease object with given ip address and return the
+		associated mac address.
+
 		@type ip: str
 		@rtype: str or None
 		@raises ValueError:
 		@raises OmapiError:
+		@raises OmapiErrorNotFound: if no lease object with the given ip
+				address could be found or the object lacks a mac address
 		@raises socket.error:
 		"""
-		msg = OmapiMessage.open("lease")
-		msg.obj.append(("ip-address", pack_ip(ip)))
+		msg = OmapiMessage.open(b"lease")
+		msg.obj.append((b"ip-address", pack_ip(ip)))
 		response = self.query_server(msg)
 		if response.opcode != OMAPI_OP_UPDATE:
 			raise OmapiErrorNotFound()
 		try:
-			return unpack_mac(dict(response.obj)["hardware-address"])
+			return unpack_mac(dict(response.obj)[b"hardware-address"])
 		except KeyError: # hardware-address
 			raise OmapiErrorNotFound()
 
